/**
    \mainpage bersicht

    \section intro Aufgabenstellung
    Der Treiber soll grundstzlich folgenden Aufgabenstellung erfllen knnen:<br><br>
    Es soll mglich sein, Daten von einer Anwendung aus auf die parallele Schnittstelle
	zu schreiben, und diese auch wieder lesen zu knnen. Um Interrupts auf der Schnittstelle
	zu erzeugen, mu der Pegel auf Pin 10 von low auf high gezogen werden. Dies kann mit 
	einem selbstgelteten Stecker erfolgen, der z.B. Pin 9 mit Pin 10 verbindet.<br><br>
	Das vorliegende Codegerst ist entsprechend zu erweitern. Smtliche Funktionen, bis auf 
	eine notwendige Hilfsfunktion sind vorgegeben und mit vielen Kommentaren versehen.

	\section extra Zu implementierende Hilfsfunktion
	Die Logik fr den Treiber soll folgende sein: Das Anwendungsprogramm schreibt Daten 
	auf das Gert. Der Treiber nimmt diese Daten entgegen und schickt sie Bit-Weise
	ber eine Datenleitung des Parallelports (oder auch 4-Bit-wesie). Diese Datenleitung ist z.B. mit 
	einem Pin des Statusregisters verbunden, Pin2 an Pin15 etwa. Sobald dieses eine Bit auf 
	Pin2 geschrieben wurde, kann es in der gleichen Funktion sofort wieder von Pin15 gelesen 
	werden. Daraufhin ist ein Interrupt zu erzeugen. Dieser endet unweigerlich mit dem Aufruf 
	unserer bereitgestellten ISR. Diese stt dann die bertragung des nchsten Zeichens an.
	Die Funktion, welche die "bertragung" der Daten und Auslsen des Interrupts ermglicht, 
	ist, wie gesagt, noch zu implementieren. Sinn und Zweck dieser bung ist das Nachbilden 
	der generellen Funktionsweise eines interrupt-gesteuerten IO-Vorgangs.
*/

#include "Driver.h"

///Wird beim erstmaligen Laden des Treibers aufgerufen. Initialisiert den Treiber.
/**
@param pDriverObject Wird vom PNP-Manager bergeben und reprsentiert den Treiber im Kernel.
@param pRegistryPath Dieser Zeiger auf einen Unicode-String entspricht dem Schlssel
	aus der	Registry fr diesen Treiber. In userem Beispiel nicht von Belang.

@return Rckgabewert ist ein NTSTATUS-Code (LONG), der Erfolg bzw. Fehlercode signalisiert.
	Die Auswertung erfolgt mittels des Makros NT_SUCCESS.
*/
NTSTATUS DriverEntry (IN PDRIVER_OBJECT pDriverObject, IN PUNICODE_STRING pRegistryPath) {

	DbgPrint ("ParpLoop: DriverEntry betreten\n");
	
	NTSTATUS status = STATUS_SUCCESS;

	//zur Einstimmung eine einfache Aufgabe...
	//es mssen folgende Funktionen des Treibers dem IO- und
	//PNP-Manager bekannt gemacht werden:
	//DriverUnload, AddDevice, DispPnp, DispatchCreate, 
	//DispatchClose, DispatchWrite, DispatchRead
	//die dafr notwendige Struktur findet sich in dem
	//Argument pDriverObject, Zeiger auf eine Struktur vom
	//Typ DRIVER_OBJECT
	//fr Funktionen wie DispatchWrite etc. existieren
	//Codes wie IRP_MJ_WRITE, die fr den Zugriff 
	//(bzw. zur korrekten Indexierung in das Array)
	//auf das Array in pDriverObject ntig sind.		
	pDriverObject->DriverUnload = DriverUnload;
	/* HIER CODE EINFGEN */ /* CODE */
	//oben genannte Funktionen entsprechend in das Array eintragen

	DbgPrint ("ParpLoop: DriverEntry erfolgreich\n");
	
	return status;
}

///Wird vom PnP-Manager aufgerufen, um den Treiber zu entladen.
/**
Der Treiber kann noch ausstehende Arbeiten ausfhren, sofern dies ntig ist.
Der IRQL-Kontext ist PASSIVE_LEVEL, so dass eben auch Speicher wieder freigegeben
werden kann.

@param pDriverObject Zeiger auf TreiberObjekt, welches vom IO-Manager verwaltet
	wird. Es ist NICHT unsere Aufgabe, dieses zu lschen. 
*/

VOID DriverUnload (IN PDRIVER_OBJECT pDriverObject) {
	DbgPrint("ParpLoop: Treiber wird entladen\n");
}

///Sobald ein neues Gert am Bus gefunden wurde, wird AddDevice aufgerufen.
 
/**
AddDevice dient dazu, den Gertestack aufzubauen. Es mu also fr
das PhysicalDeviceObject (PDO) ein entsprechendes FunctionalDeviceObject (FDO)
erzeugt werden. Auerdem ist eine "Bekanntmachung" des Gertes in Form eines 
symbolischen Gertenames erforderlich.

@param pDriverObject Das fr diesen Treiber angelegte Objekt.
@param pdo Ein Zeiger auf das PDO, welches fr den Zugriff auf die Hardware vom
	PNP-Manager angelegt wurde. WICHTIG: Fr unsere Funktion stellt sich dieses Objekt
	als PDO dar, kann aber im Kontext des Treiberstacks auch die Rolle eines
	Filterobjektes spielen, was fr uns aber vllig transparent und auch egal ist.

@return Ein Long-Wert, verpackt als NTSTATUS, welcher mittels NT_SUCCESS 
	ausgwertet werden kann.
*/
NTSTATUS AddDevice (IN PDRIVER_OBJECT pDriverObject, IN PDEVICE_OBJECT pPdo) {
	
	NTSTATUS status;
	//diese Functional Device Object werden wir anlegen
	PDEVICE_OBJECT pFdo;
	//Fr jedes "physische" Gert, das der Treiber bedient, muss dieser ein DeviceObject
	//erstellen, welches genau diesem Gert eindeutig zugeordnet ist.
	//Einem jeden solchen Device-Object kann ein Zeiger fr eine eigene Struktur mitgegeben werden.
	//Muss man z.B. mit einem Treiber zwei serielle Schnittstellen bedienen, bietet sich
	//zur Identifizierung der jeweiligen Schnittstelle diese Erweiterung an, da der Zeiger auf 
	//das jeweilige Device-Object (fast) jedem Funktionsaufruf im Treiber mit bergeben wird. 
	PDEVICE_EXTENSION pDevExt;
	
	//wir bentigen einen eindeutigen Namen fr das jeweilige Gert, der im Namensraum
	//des IO-Managers eingetragen wird
	static int deviceNumber = 0;	
	DbgPrint("ParpLoop: Aktueller Wert fr DeviceNumber = %d\n", deviceNumber);
	
	CUString devName("\\Device\\ParpLoop"); 
	devName += CUString(deviceNumber);

	//um das Gerteobjekt anzulegen, ist ein entsprechender Aufruf der
	//Funktion IoCreateDevice erforderlich
	/* HIER CODE EINFGEN */ /* CODE */	
	//IoCreateDevice, mit entsprechender berprfung des Rckgabecodes
	//und mglichem Abbruch
	
	//Wir mssen und jetzt entscheiden, welche Art der Datenbermittlung aus dem 
	//User-Adressraum wir wollen.
	//Die einfachste Mglichkeit ist, den IO-Manager den User-Puffer
	//in einen internen Puffer zu kopieren zu lassen und diesen uns zur Verfgung zu stellen.
	//Es muss dazu das Feld Flags unseres FDO mit dem entsprechenden Wert gesetzt werden	
	/* HIER CODE EINFGEN */ /* CODE */

	
	
	//Wie oben angedeutet, nutzen wir die Mglichkeit, unser DeviceObject
	//mit einem eigenen Pointer zu erweitern. 
	//Folgende Mglichkeiten sollen uns durch die eigene Struktur zur Verfgung stehen:
	//1) Schnelle Zugriff auf das zugeordnete DeviceObject
	//2) Der intern verwendete Name soll schnell auffindbar sein
	//3) Die verwendete lauffende Nummer fr dieses DeviceObject soll abrufbar sein
	//4) Der derzeitige Zustand ist gestoppt
	//5) WICHTIG: Natrlich muss der Zeiger in unserem DeviceObject auch auf
	//   die DeviceExtension gesetzt werden
	//siehe vordefinierte Struktur in Driver.h
	/* HIER CODE EINFGEN */ /* CODE */


	//Diese beiden Initialisierungen dienen der logischen Funktionsweise
	//des Treibers, Erklrung siehe weiter unten.
	pDevExt->pIntObj = NULL;
	pDevExt->InterruptExpected = FALSE;


	//Als nchste Aufgabe mssen wir erreichen, dass wir den Treiber-Stack weiter aufbauen.
	//Es ist dabei zu beachten, dass die Information auch irgendwie gespeichert werden mu.
	//Die erforderliche Voraussetzung existiert bereits!!!
	/* HIER CODE EINFGEN */ /* CODE */
	//IoAttachDeviceToDeviceStack + Zusatz
		

	//Unter Umstnden kann es erforderlich sein, auch den weiter unten
	//liegenden Treiber zu nutzen. Doch dazu mssten wir die genaue
	//Definition der DeviceExtension des unteren Treibers kennen.
	//Es soll an dieser Stelle nur als Kommentar eine mgliche 
	//Verdrahtung symbolisch implementiert werden. Die dabei notwendige
	//DeviceExtension soll als LOWER_DEVICE_EXT definiert sein. Wir selber
	//nutzen unsere DeviceExtension, um die Verdrahtung zu ermglichen.
	/* HIER CODE EINFGEN */ /* WICHTIG! auskommentieren! */


	//Nachdem die vorherigen Initialisierungen erfolgreich waren, ist
	//es ntig, den DeviceNamen zu erzeugen und zu verffentlichen, indem wir einen
	//Symbolischen Verbindungsnamen bilden. Der Namenraum des IO-Managers
	//hierfr beginnt mit \??. Siehe erste bung.
	//Es ist dabei darauf zu achten, entsprechend den verschiedenen mglichen 
	//DeviceObjects eindeutige Namen zu vergeben und auerdem die Informationen
	//fr sptere Zugriffe wieder zu speichern.
	/* HIER CODE EINFGEN */ /* CODE */
	//Erzeugen des Namens + IoCreateSymbolicLink + Statusprfung
	//WICHTIG: Im Fehlerfall ist auch das DeviceObject wieder zu entfernen!


	//An dieser Stelle ist ein wenig Vorarbeit notwendig. Unser Treiber wird
	//mit Interrupts arbeiten. Wir mssen also Windows eine Mglichkeit
	//geben, uns im Falle des Anliegen eines Interrupts zu benachrichtigen. 
	//Die dazu notwendige Funktion heit in unserem Fall Isr (InterruptServiceRoutine).
	//Damit ist es allerdings nicht getan. Wir mssen auch die Mglichkeit bieten, 
	//Aufgaben, welche nach der Interruptbehandlung anfallen, zu bearbeiten.
	//Fr diesen Zweck implementiert der Treiber eine Methode: DpcForIsr,
	//Deferred Prcedure Call for Interrupt Service Routine, ein verzgerter
	//Funktionsaufruf also, der auf einem niedrigeren IRQL (siehe Vorlesung) 
	//ausgefhrt wird.
	/* HIER CODE EINFGEN */ /* CODE */
	//IoInitializeDpcRequest
	
	//Der folgende Aufruf lscht dieses Flag fr unser DeviceObject. Dieser
	//Aufruf ist bindend fr Plug'nPlay Treiber. 
	//Abschnitt 3.2.3 gibt einen kurzen berblick ber das Flag-Feld von DEVICE_OBJECT.
    pFdo->Flags &= ~DO_DEVICE_INITIALIZING;

	//Da wir das DeviceObject erzeugt haben, mssen wir fr den nchsten Zugriff
	//noch die laufende Nummer erhhen.
	deviceNumber++;
	
	return STATUS_SUCCESS;
}


///Wird vom PnP-Manager aufgerufen, um entsprechende PnP-Codes zu bergeben. 
/**
@param pDo Ist der Zeiger auf das DeviceObject, welches wir in AddDevice
	angelegt haben.
@param pIrp Zeiger auf das entsprechende IO-Request-Packet. Dieses 
	durchwandert den gesamten Stack, anschaulich gesagt, zuerst von 
	oben nach unten und wieder zurck.
@return Ein Long-Wert, verpackt als NTSTATUS, der mittels NT_SUCCESS berprft 
	ausgewertet werden kann.
*/
NTSTATUS DispPnp(IN PDEVICE_OBJECT pDo, IN PIRP pIrp) {
		
	PIO_STACK_LOCATION pIrpStack;
	//Da unser Treiber Bestandteil eines 
	//Stacks ist, mssen wir erst einen gltigen Zeiger in den IO-Stack des 
	//IRP bekommen.	
	/* HIER CODE EINFGEN */ /* CODE */
	//unter IRP in der DDK-Help steht die entsprechende Funktion und auerdem
	//eine kurze lesenswerte Erluterung

		
	DbgPrint("ParpLoop: PNP KontrollCode: %d\n", pIrpStack->MinorFunction);

	
	switch (pIrpStack->MinorFunction) {
	case IRP_MN_START_DEVICE:
		return HandleStartDevice(pDo, pIrp );
	case IRP_MN_STOP_DEVICE:
		return HandleStopDevice( pDo, pIrp );
	case IRP_MN_REMOVE_DEVICE:
		return HandleRemoveDevice( pDo, pIrp );
	default:
		//PassDownPnP ist eine von uns erstellte Funktion, die
		//sehr hufig verwenet wird, um PnP-Codes an die
		//unten liegenden Treiberschichten zu bergeben.
		return PassDownPnP(pDo, pIrp);
	}
}


///Zentrale Funktion, die die Weiterleitung nach IO-Request-Packets durch den Treiberstack ermglicht.
/**
@param pDo Zeiger auf von uns in AddDevice angelegtes DeviceObject.
@param pIrp Zeiger auf entsprechendes IRP.
@return Long-Wert, verpackt als NTSTATUS. Kann mittels NT_SUCCESS ausgewertet werden.
*/
NTSTATUS PassDownPnP(IN PDEVICE_OBJECT pDo, IN PIRP pIrp) {
	
	//Diese Funktion soll eine Weiterleitung von IRPs "nach unten" in den Treiberstack
	//ermglichen. Dafr sind meherer Schritte erforderlich. Wir mssen zuerst die
	//Daten des IRP-IO-Stacks in den Stack fr den weiter unten liegenden Treiber kopieren,
	//um diesen dann mit den gesetzten Werten aufzurufen. Es gibt hierfr 2 Mglichkeiten.
	//1) Wir knnen die Werte kopieren, und den Zustand des IRP fr unsere Zwecke als
	//	 in Bearbeitung markieren
	//2) Wir berspringen einfach unseren eigenen Treiber und reichen das IRP
	//	 direkt an die nchste Treiberschicht weiter
	//Wo liegt der Unterschied?! Im ersten Fall erwarten wir "die Rckkehr" des IRP
	//durch den Stack nach oben zurck. Wurde z.B. eine IO-Anforderung angestossen, 
	//mssen wir u.U. zuerst den eigentlichen Hardwarezugriff von einer unteren Treiberschicht
	//erledigen lassen, um dann selbst das Ergebnis mglicherweise zu filtern,
	//und an die Anwendung zurckzugeben. Im zweiten Fall interessieren wir uns nicht
	//fr das Ergebnis der unten liegenden Treiberschichten, drfen in diesem Fall aber
	//auch keine weitere Verfgung ber das IRP haben im Sinne eines Aufrufes von 
	//IoSetCompletionRoutine etc.

	/* HIER CODE EINFGEN */ /* CODE */
	//diese Treiberschicht berspringen + weiter unten liegenden Treiber aufrufen
	//KEINE Completion-Routine setzen!!!
	//IoSkipCurrentStackLocation, IoCallDriver
	//ACHTUNG: Trotz dessen wir uns nicht weiter fr den unmittelbaren Erfolg des Treiberaufrufes
	//interessieren, mssen wir doch den PnP-Manager ber den Erfolg wissen lassen...
}


///Wird aus dem Treiber heraus aufgerufen.
/**
Um eine bessere Verteilung der Funktionalitt und Wartbarkeit des Codes zu erreichen, 
wird der KontrollCode IRP_MN_START_DEVICE auf diese Funktion abgebildet. 
@param pDo Zeiger auf von uns erzeugtes DeviceObject.
@param pIrp Zeiger auf entsprechendes IRP
@return Long-Wert, verpackt als NTSTATUS. Kann mittels NT_SUCCESS ausgewertet werden.
@see DispPnP
*/
NTSTATUS HandleStartDevice(IN PDEVICE_OBJECT pDo, IN PIRP pIrp) {
	
	PIO_STACK_LOCATION pIrpStack;

	//Wir mssen zuerst wieder die korrekte Position innerhalb des IO-Stacks
	//in dem IRP extrahieren, siehe DispPnP
	/* HIER CODE EINFGEN */ /* CODE */

	
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pDo->DeviceExtension;

	DbgPrint("ParpLoop: Folgendes Gert soll gestartet werden: #%d\n", pDevExt->DeviceNumber + 1);


	//die folgenden Ausfhrungen sind Bestandteil des PnP und des generellen
	//Treibermodells von Windows. Es 
	PCM_RESOURCE_LIST pResourceList;
	PCM_FULL_RESOURCE_DESCRIPTOR pFullDescriptor;
	PCM_PARTIAL_RESOURCE_LIST pPartialList;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR pPartialDescriptor;
	int i;
	
	NTSTATUS status;

	pResourceList =	pIrpStack->Parameters.StartDevice.AllocatedResourcesTranslated;
	pFullDescriptor = pResourceList->List;
	pPartialList = &pFullDescriptor->PartialResourceList;
	
	for (i=0; i<(int)pPartialList->Count; i++) {
		
		pPartialDescriptor = &pPartialList->PartialDescriptors[i];
		switch (pPartialDescriptor->Type) {
		case CmResourceTypeInterrupt:
			pDevExt->IRQL = (KIRQL) pPartialDescriptor->u.Interrupt.Level;
			pDevExt->Vector = pPartialDescriptor->u.Interrupt.Vector;
			pDevExt->Affinity = pPartialDescriptor->u.Interrupt.Affinity;
			
			DbgPrint("ParpLoop: Belegte Interrupt-Ressource: "\
				"IRQ=%d Vector=0x%03X Affinity=%X\n", \
				pDevExt->IRQL, pDevExt->Vector, pDevExt->Affinity);
			break;
		case CmResourceTypeDma:
			break;
		case CmResourceTypePort:
			pDevExt->portBase = (PUCHAR) pPartialDescriptor->u.Port.Start.LowPart;
			pDevExt->portLength = pPartialDescriptor->u.Port.Length;

			DbgPrint("ParpLoop: Reservieren von Ports: Base=%X Len=%d\n", \
				pDevExt->portBase, pDevExt->portLength);
			break;
		case CmResourceTypeMemory:
			break;
		}
	}

	//Es ist erforderlich, zu berprfen, ob der Schnittstelle ein
	//Interrupt zugeordnet werden konnte.
	if (pDevExt->IRQL == 0 || pDevExt->portBase == 0)
		return STATUS_BIOS_FAILED_TO_CONNECT_INTERRUPT;

    // Anlegen und Verbinden eines Interrupt-Objekts
	status = IoConnectInterrupt(
			&pDevExt->pIntObj,	//Interrupt-Objekt 
			Isr,				//unsere ISR
			pDevExt,			//dieser Zeiger wird dann der ISR genauso mit bergeben
			NULL,				//wir brauchen kein Spin Lock
			pDevExt->Vector,	// 
			pDevExt->IRQL,		// 
			pDevExt->IRQL,		// 
			Latched,			//flankengesteuert
			TRUE,				//kann der Interruptvektor gemeinsam genutzt werden
			pDevExt->Affinity,	//gibt an, auf welchen Prozessoren dieser Interrupt auftreten kann
			FALSE);				//soll das FloatingPoint-Register gespeichert werden; mu auf X86 immer FALSE sein
	
	if (!NT_SUCCESS(status)) {
		DbgPrint("ParpLoop: Interrupt konnte nicht zugeordnet werden%X\n", status);
		return status;
	}
	

	DbgPrint("ParpLoop: Interrupt zugeordnet\n");

	pDevExt->state = Started;

	return PassDownPnP(pDo, pIrp);
}

///Aufruf aus dem Treiber heraus.
/**
@param pDo Von uns erzeugtes DeviceObject
@param pIrp Zeiger auf entsprechendes IRP
@return Long-Wert, verpackt als NTSTATUS. Kann mittels NT_SUCCESS ausgewertet werden.
@see HandleStartDevice
@see DispPnp
*/
NTSTATUS HandleStopDevice(IN PDEVICE_OBJECT pDo, IN PIRP pIrp) {

	DbgPrint("ParpLoop: Gert wird angehalten\n");

	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION)
		pDo->DeviceExtension;

	//an dieser Stelle wird das Gert gestoppt, wir mssen dafr 
	//alle unmittelbaren Systemressource freigeben, v.a.
	//mssen wir das Gert vom Interrupt lsen, soweit wir diesen berhaupt 
	//schon belegt haben
	/* HIER CODE EINFGEN */ /* CODE */
	//IoDisconnectInterrupt
	

	pDevExt->state = Stopped;

	//nachdem unsere Treiber vom Interrupt gelst wurde, mssen wir noch dafr sorgen, 
	//da die unteren Treiberschichten informiert werden und auerdem auch der PnP-Manager
	//bescheid bekommt...
	/* HIER CODE EINFGEN */ /* CODE */
}

///Aufruf aus dem Treiber heraus.
/**
@param pDo Von uns erzeugtes DeviceObject
@param pIrp Zeiger auf entsprechendes IRP
@return Long-Wert, verpackt als NTSTATUS. Kann mittels NT_SUCCESS ausgewertet werden.
@see HandleStartDevice
@see DispPnp
*/
NTSTATUS HandleRemoveDevice(IN PDEVICE_OBJECT pDo, IN PIRP pIrp ) {

	DbgPrint("ParpLoop: Gert wird entfernt\n");

	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pDo->DeviceExtension;

	//Wir mssen dafr sorgen, da auch im Fall der Flle der Treiber
	//ordnungsgem den Interrupt wieder freigibt
	if (pDevExt->state == Started)
		/* HIER CODE EINFGEN */ /* CODE */
		//IoDisconnectInterrupt
	
	pDevExt->state = Stopped;

	
	//Da wir das Gert entfernen und spter auch das zugehrige DeviceObject, 
	//mssen wir natrlich auch den symbolischen Link lschen
	/* HIER CODE EINFGEN */ /* CODE */
	//IoDeleteSymbolicLink + DeviceObject lschen + Debugmeldung ausgeben
	
	
	pDevExt->state = Removed;
	
	//als Standardaufgabe mssen wir natrlich wieder die weiter unten liegenden Treiberschichten
	//informieren und das Ergebnis auch den PnP-Manager wissen lassen
	/* HIER CODE EINFGEN */ /* CODE */
}

///Wird direkt vom IO-Manager aufgerufen, um den Win32-Funktionsaufruf CreateHandle zu bearbeiten.
/**
@param pDevObj Von uns erzeugtes DeviceObject, welches der IO-Manager dem Funktionsaufruf
	mit bergibt. WIr knnen damit das angesprochene Gert eindeutig zuordnen.
@param pIrp Zugehriger Zeiger auf ein IRP.

@return Long-Wert, verpackt als NTSTATUS. Kann mittels NT_SUCCESS ausgewertet werden.
*/

NTSTATUS DispatchCreate (IN PDEVICE_OBJECT	pDevObj, IN PIRP pIrp) {

	DbgPrint("ParpLoop: DispatchCreate wurde aufgerufen\n");

	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION)
		pDevObj->DeviceExtension;
		
	NTSTATUS status = STATUS_SUCCESS;
	
	//An dieser Stelle knnen Fehler auftreten. Es soll das 
	//von uns bediente Gert geffnet werden. Allerdings kann es sein, 
	//da das Gert nicht gestartet wurde. Dies ist mit einem Fehlerfall zu quittieren.
	if (pDevExt->state != Started)
		status = STATUS_DEVICE_REMOVED;
	
	pIrp->IoStatus.Status = status;
	pIrp->IoStatus.Information = 0;	// no bytes xfered
	
	//Nachdem wir wissen, dass das Gert gestartet wurde, mssen 
	//wir das IRP an den IO-Manager zurckgaben. Wir mssen
	//allerdings auch die Bearbeitung als abgeschlossen erklren.
	/* HIER CODE EINFGEN */ /* CODE */
	//IoCompleteRequest, es ist darauf zu achten, den richtigen Parameter zu setzen.
	//Die DDK-Hilfe liefert hierfr eine kurze Erklrung

	return status;
}

///Wird direkt vom IO-Manager aufgerufen, um den Win32-Funktionsaufruf CloseHandle zu bearbeiten.
/**
@param pDevObj Zeiger auf das entsprechende DeviceObject, welches uns eine eindeutige
	Identifikation des Gertes ermglicht.
@param pIrp Zeiger auf entsprechendes IRP
@return Long-Wert, verpackt als NTSTATUS. Kann mittels NT_SUCCESS ausgewertet werden.
*/
NTSTATUS DispatchClose (IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp) {

	DbgPrint("ParpLoop: DispatchClose wurde aufgerufen\n");
	
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pDevObj->DeviceExtension;
	
	//an dieser Stelle ist dafr zu sorgen, da der 
	//Puffer, falls belegt, wieder freigegeben wird.
	//Dies mu mit der entsprechend KERNEL-Funktion geschehen.
	/* HIER CODE EINFGEN */ /* CODE */
	//Puffer freigeben + Informationen in der DeviceExtension setzen.
	
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;
	
	//Es ist wieder erforderlich, das IRP als bearbeitet zu erklren. Dabei
	//ist auch wieder auf die richtige Parametrisierung des Funktionsaufrufes 
	//zu achten
	/* HIER CODE EINFGEN */ /* CODE */
	
	return STATUS_SUCCESS;
}

///Wird aufgerufen, wenn die Bearbeitung eines IRP abgebrochen wurde.
/**
@param pDevObj Zeiger auf das entsprechende DeviceObject, welches uns eine eindeutige
	Identifikation des Gertes ermglicht.
@param pIrp Zeiger auf entsprechendes IRP
*/
VOID DispatchCancel (IN PDEVICE_OBJECT	pDevObj, IN PIRP pIrp) {

	DbgPrint("ParpLoop: IRP-Bearbeitung wurde abgebrochen\n");
	
	//Wir mssen an dieser Stelle dafr sorgen, da diese IRP verworfen
	//wird, aber gleichzeitig bereits sich in der Warteschlange befindliche IRPs
	//bearbeitet werden knnen
	/* HIER CODE EINFGEN */ /* CODE */
	//Status des IRP auf Cancel setzen, siehe ntStatus.h des DDK +
	//IRP als abgeschlossen markieren +
	//nchstes Packet aus der Schlange holen und zur Bearbeitung freigeben, IoStartNextPacket
}


///Wird fr die Win32-Funktion WriteFile aufgerufen
/**
@param pDevObj pDevObj Zeiger auf das entsprechende DeviceObject, welches uns eine eindeutige
	Identifikation des Gertes ermglicht.
@param pIrp pIrp Zeiger auf entsprechendes IRP
@return Long-Wert, verpackt als NTSTATUS. Wird vom IO-Manager ausgewertet.
*/
NTSTATUS DispatchWrite (IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp) {

	DbgPrint("ParpLoop: Daten werden auf das Gert geschrieben\n");
	
	//Wir mssen jetzt erreichen, da alle Treiberschichten nach unten hin durchlaufen
	//werden. Es ist dafr erforderlich, das derzeitige IRP als in Bearbeitung zu markieren
	//und den eigentlichen IO-Vorgang zu starten	
	IoMarkIrpPending( pIrp );	
	/* HIER CODE EINFGEN */ /* CODE */
	//IoStartPacket
	
	
	//Wir mssen jetzt aber auch den IO-Manager informieren, dass
	//der angeforderte IO-Vorgang bearbeitet wird
	/* HIER CODE EINFGEN */ /* CODE */
	//siehe ntStatus.h des DDK
}

///Wird fr die Win32-Funktion ReadFile aufgerufen.
/**
@param pDevObj pDevObj Zeiger auf das entsprechende DeviceObject, welches uns eine eindeutige
	Identifikation des Gertes ermglicht.
@param pIrp pIrp Zeiger auf entsprechendes IRP
@return Long-Wert, verpackt als NTSTATUS. Wird vom IO-Manager ausgewertet.
*/
NTSTATUS DispatchRead (IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp) {
	
	DbgPrint("ParpLoop: Daten werden vom Gert gelesen\n");
	
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pDevObj->DeviceExtension;
	
	NTSTATUS status = STATUS_SUCCESS;
	
	PVOID userBuffer;
	ULONG toTransfer;
	
	PIO_STACK_LOCATION pIrpStack;
	//Es ist an dieser Stelle die korrekte Stack-Position
	//im IRP zu ermitteln. Siehe DispPnP
	/* HIER CODE EINFGEN */ /* CODE */

	
	//Das IRP liefert uns smtliche Informationen, die wir bentigen,
	//um die geforderte Menge an Bytes zu bertragen.
	//Unter dem Stichwort IO_STACK_LOCATION liefert die
	//DDK-Hilfe eine Beschreibung dieser Struktur. Weiterhin liefert
	//die Suche nach IRP_MJ_READ die noch ntigen Informationen.
	/* HIER CODE EINFGEN */ /* CODE */
	//toTransfer setzen +
	//userBuffer auf korrekte Adresse setzen +
	//sicherstellen, da toTransfer die Gre des 
	//Puffers nicht berschreitet

	
	
	//Es ist jetzt dafr zu sorgen, dass die Daten aus dem 
	//Kernel-Puffer auch zur Anwendung gelangen.
	//Es ist dafr die KERNEL-Funktion zum 
	//Speicherbereich-Kopieren zu verwenden
	//Weiterhin mu der Speicher, den der Treiber
	//allokiert hat, wieder freigegeben werden, widerrum ist die
	//KERNEL-Funktion zu verwenden
	/* HIER CODE EINFGEN */ /* CODE */	
	//Pufferinhalt kopieren + freigeben + interne Informationen anpassen
	

	//Da es sich wieder um einen IRP-gesteuerten Vorgang handelt,
	//mu dieses natrlich wieder als abgearbeitet markiert werden.
	//Es ist dabei darauf zu achten, da auch die Anzahl der bertragenen
	//Zeichen in irgendeiner Form dem IO-Manager mitgeteilt werden mu.
	//(Hilfe unter IRP_MJ_READ)
	// IRP komplettieren
	/* HIER CODE EINFGEN */ /* CODE */
	//status setzen + Anzahl der bertragenen Bytes signalisieren +
	//Bearbeitung des IRP als abgeschlossen markieren
	
	
	return status;
}


///Eigentlicher Einsprungpunkt, um einen aufgetretenen Interrupt zu bearbeiten.
/**
@param pIntObj Zeiger auf das unserem Device zugeordnete InterruptObject. Dieses
	wurde in HandleStartDevice angelegt.
@param pServiceContext Zeiger auf eine von uns bereitgestellte Struktur, 
	der DeviceExtension. Auf diese Art und Weise knnen wir Rckschlsse auf
	das entsprechende DeviceObject machen.
@return Der Rckgabewert ist TRUE, falls die ISR den Interrupt nicht behandelt,
	weil er nicht von der parallelen Schnittstelle stammt. FALSE andernfalls. Der Kernel
	wird dann die nchste ISR aufrufen, die ebenfalls auf diesen Interrupt 
	abgebildet worden ist.
*/
BOOLEAN Isr (IN PKINTERRUPT pIntObj, IN PVOID pServiceContext) {

	DbgPrint ("Interrupt eingetreten\n");
	
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pServiceContext;
	PDEVICE_OBJECT pDevObj = pDevExt->pDevice;
	PIRP pIrp = pDevObj->CurrentIrp;

	//Wir mssen jetzt berprfen, ob der Interrupt von
	//unserer Schnittstelle stammt. Sollte dem nicht so sein, mssen wir FALSE
	//zurckgeben.
	/* HIER CODE EINFGEN */ /* CODE */
	//Statusregister der parallelen Schnittstelle auslesen und
	//auf anstehenden Interrupt hin berprfen, ggf. Rcksprung
	

    //Sind wir an dieser Stelle angekommen, liegt ein Interrupt an. 
	//Das Bit fr den anstehenden Interrupt mu jetzt gelscht werden.
	/* HIER CODE EINFGEN */ /* CODE */ 

	
	//Wird das Gert mit einem Interruptobjekt verbunden (in Funktion HandleStartDevice),
	//werden auch Interrupts erzeugt. Diese Interrupts haben allerdings nichts
	//mit unserer Datenbertragung zu tun. Wir mssen also eine Logik finden, welche
	//es uns erlaubt, diese "unfreiwilligen" Interrupts zu umgehen, 
	//dafr existiert das Feld InterruptExpected in unserer DeviceExtension.	
	if (!pDevExt->InterruptExpected)
		return TRUE;
	
	
	pDevExt->InterruptExpected = FALSE;

	//ACHTUNG: Zur Erklrung der Ablauflogik bitte am Anfang dieser Datei nachlesen.
	//An dieser Stelle mssen wir erreichen, da das nchste Bit des anstehenden Zeichens
	//bertragen wird. Es ist dabei darauf zu achten, da das IRP auf irgendeine Art und
	//Weise abgeschlossen werden mu, sobald keine Zeichen mehr im Eingangspuffer
	//zur Verfgung stehen
	/* HIER CODE EINFGEN */ /* CODE */
	//Zeichen bertragen + bei Ende der bertragung einen verzgerten
	//Funktionsaufruf beantragen, Stichwort DPC
	

	return TRUE;
}


///Funktion fr den eigentlichen Datentransfer.
/**
@param pDevObj Zeiger auf das von uns erzeugte DeviceObject
@param pIrp Zeiger auf entsprechendes IRP.
*/

VOID StartIo(IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp) {

	DbgPrint("ParpLoop: Funktion StartIo aufgerufen\n");
	
	PIO_STACK_LOCATION	pIrpStack;
	//korrekte StackPosition innerhalb des IRP herausfinden. Siehe hierfr
	//Erklrung in DispPnP
	/* HIER CODE EINFGEN */ /* CODE */
		
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pDevObj->DeviceExtension;
	
	PUCHAR userBuffer;
	ULONG toTransfer;

	//Wir verarbeiten an dieser Stelle definitiv eine Schreibanforderung. Entsprechend
	//ist der Switch-Block aufzubauen. Die Programm-Logik ist in dieser Datei am 
	//Anfang als Kommentar beschrieben.
	switch( pIrpStack->MajorFunction ) {
		
	case 0: /* CODE */
		
		/* HIER CODE EINFGEN */ /* CODE */
		//Korrekten FunktionsCode verarbeiten +
		//Puffer entsprechend allokieren, bzw. vorher freigeben +
		//Anzahl der zu bertragenden Zeichen aus dem
		//IRP-Stack ermitteln + 
		//zu implementierende (!!!) Hilfsfunktion aufrufen +
		//WICHTIG: diese Funktion mu unbedingt
		//mit der ISR synchronisiert werden im Sinne
		//eines gegenseitigen Ausschlusses!	-> KeSynchronizeExecution
		break;		
	
	default:
		//sollten wir an dieser Stelle irgendeinen anderen
		//FunktionsCode als den oben angegebenen erhalten, 
		//so ist dieses IRP zu verwerfen und dessen Behandlung
		//als abgeschlossen zu markieren
		/* HIER CODE EINFGEN */ /* CODE */
		//Status fr IRP setzen + 
		//Anzahl der bertragenen Zeichen auf 0 setzen +
		//IRP abschlieen + 
		//nchstes Packet starten
		break;
	}
}

///Wird vom IO-Manager als DPC aufgerufen
/**
Ist ein Interrupt aufgetreten, wird die entsprechende ISR aufgerufen. Da diese
entsprechend kurz gehalten werden mu, knnen ausstehende
Arbeiten in eine eigene Funktion ausgelagert werden, in diesem Fall
DpcForIsr. Diese wird dann auf einem niedrigeren IRQL ausgefhrt. Es
ist dabei darauf zu achten, da ein mehrmaliges Einfgen eines DPC-Aufrufes 
in die Windows-interne Warteschlange in diesem einfachen Fall nicht mglich ist.
Der Aufruf dafr kann zwar mehrfach erfolgen, die Ausfhrung erfolgt jedoch
nur einmal. Wird jedoch eine strenge 1:1 Abbildung von IoRequestDpc auf DpcForIsr
gefordert, mu man einen eigenen Mechanismus implementieren.

@param pDpc Ein Zeiger auf DPC-Objekt, siehe hierzu wdm.h
@param pDevObj Zeiger auf unser DeviceObject
@param pIrp Zeiger auf entsprechendes IRP
@param pContext Zeiger auf unsere eigene Struktur, in diesem Fall unsere DeviceExtension
*/
VOID DpcForIsr(IN PKDPC pDpc, IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp, IN PVOID pContext)
{
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pContext;	

	DbgPrint("ParpLoop: In DpcForIsr, Anzahl bertragener Zeichen = %d\n", pDevExt->xferCount);

	
	//Wir mssen jetzt die ntigen Information im IPR setzen
	//es ist anzugeben, ob die Daten erfolgreich bertragen werden
	//konnten und vor allem, wie viele!
	/* HIER CODE EINFGEN */ /* CODE */
	//Anzahl bertragener Zeichen kenntlich machen + 
	//entsprechenden Erfolgsstatus setzen



	//Wir mssen dem IO-Manager mitteilen, da dieses IRP abgearbeitet wurde
	//Es ist zu untersuchen, ob die Bearbeitung erfolgreich war oder nicht.
	//Windows bietet fr den Abschlu von IO-Request die Mglichkeit,
	//den ausfhrenden Proze eine Sonderbehandlung angedeihen zu lassen
	//im Sinne einer Priorittenanhebung. Siehe wdm.h
	/* HIER CODE EINFGEN */ /* CODE */
	//IRP als beendet markieren, dabei auf korrekte Parametrisierung achten
	


	//Dieses IRP wurde abgearbeitet. Es knnen allerdings die nchsten IRPs
	//bereits in der Warteschlange stehen. Wir mssen dem IO-Manager mitteilen,
	//da er das nchste Packet schicken kann.
	/* HIER CODE EINFGEN */ /* CODE */
	//IoStartNextPacket
}
