extern "C" {
#include <WDM.h>
}
#include "Unicode.h"

enum DRIVER_STATE {Stopped, Started, Removed};

typedef struct _DEVICE_EXTENSION {
	PDEVICE_OBJECT pDevice;
	PDEVICE_OBJECT pLowerDevice;
	ULONG DeviceNumber;
	CUString ustrDeviceName;	//interner Name
	CUString ustrSymLinkName;	//symbolischer Verbindungsname
	PUCHAR deviceBuffer;		//temporrer Puffer
	ULONG deviceBufferSize;
	ULONG xferCount;			//Bytes transferiert
	ULONG maxXferCount;			//Bytes insgesamt zu transferieren
	ULONG nextBit;				//nchstes zu bertragenes Bit des aktuellen Zeichens
	PUCHAR portBase;			//Adresse des I/O-Registers
	ULONG  portLength;
	KIRQL IRQL;					//Irq fr Parallelschnittstelle
	ULONG Vector;
	KAFFINITY Affinity;
	PKINTERRUPT pIntObj;	    //Interrupt-Objekt
	BOOLEAN InterruptExpected;	//TRUE, wenn der Treiber einen Interrupt erwartet
	DRIVER_STATE state;		    //aktueller Zustand des Treibers
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;



extern "C" NTSTATUS DriverEntry (PDRIVER_OBJECT pDriverObject, \
	PUNICODE_STRING pRegistryPath);
static VOID DriverUnload (IN PDRIVER_OBJECT	pDriverObject);


static NTSTATUS DispatchCreate (IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp);
static NTSTATUS DispatchClose (IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp);
static NTSTATUS DispatchWrite (IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp);
static NTSTATUS DispatchRead (IN PDEVICE_OBJECT	pDevObj, IN PIRP pIrp);


static VOID StartIo(IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp);
static BOOLEAN Isr (IN PKINTERRUPT pIntObj, IN PVOID pServiceContext);
static VOID DpcForIsr(IN PKDPC pDpc, IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp, IN PVOID pContext);


static NTSTATUS AddDevice (IN PDRIVER_OBJECT pDriverObject, IN PDEVICE_OBJECT pPdo);

static NTSTATUS DispPnp(IN PDEVICE_OBJECT pDO, IN PIRP pIrp);
static NTSTATUS PassDownPnP(IN PDEVICE_OBJECT pDO, IN PIRP pIrp );

static NTSTATUS HandleStartDevice(IN PDEVICE_OBJECT pDO, IN PIRP pIrp);
static NTSTATUS HandleStopDevice(IN PDEVICE_OBJECT pDO, IN PIRP pIrp );

static NTSTATUS HandleRemoveDevice(IN PDEVICE_OBJECT pDO, IN PIRP pIrp);




//Definitionen den Offset des jeweiligen Ports
#define DATA		0
#define STATUS		1
#define CONTROL		2


//Die Makros sind folgendermaen zu verstehen:
//W_CONTROL_ACK_1
//W fr schreiben
//CONTROL in welches Register bitteschn?
//ACK dort kommt die Leitung "wieder raus", als auf Pin10
//1 fr high, 0 fr low; diese Werte kommen dann am Ausgang an
//Alles in allem:
//folgender Funktionsaufruf steht im Quelltext:
//WriteControl (pDevExt, W_CONTROL_ACK_1);
//fhrt dazu, dass enstprechend die Leitung des Parallelports
//beschrieben wird, die auf Pin10 endet, dabei wird der Pegel auf high gesetzt.
//Die 1 und 0 am Ende ist daher wichtig, da einige Eingnge bei der parallelen
//Schnittstelle invertiert werden
#define W_CONTROL_ACK_1	0x4 //dient der Interrupterzeugung
#define W_CONTROL_INT_1 0x10 //Interrupt aktivieren

//Fr die Datenbertragung kann im einfachsten Fall eine Datenleitung 
//verwendet werden, Schreiben auf das Daten-Register bertrgt 
//Bit 0 an das Bit 3 des Statusregisters
//es knnen allerdings auch max. 4 Bits auf einmal bertragen werden, 
//sofern man den Loopback-Stecker verwendet, was allerdings die Sache 
//erschwert. Das erste Bit mu an das Datenregister gehen. Die 
//anderen 3 Bit an das Control-Register. Dabei ist allerdings 
//die Invertierung zu beachten sowie das genderte Layout der Bits.
//Bsp.: Bit 0 an Data -> als Bit 3 des Statusregisters lesen
//		Bit 1,2,3 als Bit 0,1,3 an Control-Register -> 
//			als Bit 4,5,7 des Statusregisters lesen
//ACHTUNG: Bit 0,1 am Control-Register werden am Ausgang invertiert

//Folgende Makros dienen dem Zugriff auf die Register des Parallelports
#define ReadStatus(pDevExt)\
	(READ_PORT_UCHAR (pDevExt->portBase + STATUS_REG ))

#define ReadControl(pDevExt)\
	(READ_PORT_UCHAR(pDevExt->PortBase + CONTROL_REG ))

#define WriteControl(pDevExt, data)\
	(WRITE_PORT_UCHAR(pDevExt->portBase + CONTROL_REG, data))

#define WriteData(pDevExt, data)\
	(WRITE_PORT_UCHAR(pDevExt->portBase + DATA_REG, data))
